#include "maximize.hpp"
#include "sort.hpp"
#include "stdlib.h"

// The naive inner-product algorithm.
Scalar naive(Scalar* va, Scalar* vb, int N, int& nread)
{
  Scalar max = va[0] PLUS vb[0];

  for (int i = 1; i < N; i++)
  {
    Scalar t = va[i] PLUS vb[i];
    if (t COMP max)
      max = t;
  }

  nread = N;
  return max;
}

Scalar maximize(Scalar* va, Scalar* vb, int* pa, int* pb, int* painv, int* pbinv, int& nread)
{
  int start = 0;
  int pa0 = pa[0];
  int pb0 = pb[0];
  int endb = pbinv[pa0];
  int A,B,A1;
  Scalar t;

  Scalar max = va[pa0] PLUS vb[pa0];
  if ((t = va[pb0] PLUS vb[pb0]) COMP max) max = t;

  while (start < endb)
  {
    start ++;
    A = pa[start];
    B = pb[start];

    if (!(va[pa[start]] PLUS vb[pb[start]] COMP max)) break; // This condition is used in our JMLR paper, but not in the AISTATS or ICML papers.
    if ((t = va[A] PLUS vb[A]) COMP max) max = t;
    if ((t = va[B] PLUS vb[B]) COMP max) max = t;
    if ((A1 = pbinv[A]) < endb) endb = A1;
  }
  
  nread = 2*(start + 1);

  return max;
}


/** Compute the permutation function that sorts a list. */
void permfunction(Scalar* a, int* pa, int* painv, int N)
{
  for (int i = 0; i < N; i++)
    pa[i] = i;

  Scalar* a2 = new Scalar[N];
  memcpy(a2, a, N * sizeof(Scalar));

  MedianHybridQuickSort<Scalar> (a2, N, pa);

  for (int i = 0; i < N; i++)
    painv[pa[i]] = i;
  delete[] a2;
}

void computemarginal(Scalar* messagea, Scalar* messageb, Scalar** prior, int** Pa, int** Painv, Scalar* res, int N, int method, long long& nread)
{
  /// Computes message_a[a] + max_b message_b[b] + prior[a,b]
  int* pb = new int[N];
  int* pbinv = new int[N];

  // We are maximizing over b, so compute the permutation function for the message over b.
  if (method != NAIVE) permfunction(messageb, pb, pbinv, N);

  //#pragma omp parallel for
  for (int i = 0; i < N; i++)
  {
    Scalar max = 0;
    int nr = -1;
    if (method == NAIVE)
      max = naive(prior[i], messageb, N, nr);
    else if (method == FAST)
      max = maximize(prior[i], messageb, Pa[i], pb, Painv[i], pbinv, nr);
    else
    {
      printf("Unsupported method: %d\n", method);
      exit(1);
    }
    res[i] = messagea[i] PLUS max;
    nread += nr;
  }

  delete[] pb;
  delete[] pbinv;
}
