#ifndef POTENTIAL_HPP
#define POTENTIAL_HPP

#include "common.hpp"
#include "maximize.hpp"

class potential
{
  public:
    Scalar** pot;
    Scalar** pottrans;

    Scalar* pota;
    Scalar* potb;

    int** Pa;
    int** Pb;
    int** Painv;
    int** Pbinv;

    map<int, Scalar*> messagesa;
    map<int, Scalar*> messagesb;

    int N;
    int method;
    int donotdelete;

    potential(Scalar** pot, Scalar* pota, Scalar* potb, int N, int method) :
      pot(pot), pota(pota), potb(potb), N(N), method(method)
    {
      // Compute the transpose of the potential, which we shall use depending on which direction the message is sent.
      pottrans = new Scalar*[N];
      Pa = new int*[N];
      Pb = new int*[N];
      Painv = new int*[N];
      Pbinv = new int*[N];

      for (int i = 0; i < N; i++)
      {
        pottrans[i] = new Scalar[N];
        Pa[i] = new int[N];
        Pb[i] = new int[N];
        Painv[i] = new int[N];
        Pbinv[i] = new int[N];

        for (int j = 0; j < N; j++)
          pottrans[i][j] = pot[j][i];
      }

      if (method != NAIVE)
      {
        for (int i = 0; i < N; i++)
        { // Compute permutations that sort each row/column.
          permfunction(pot[i], Pa[i], Painv[i], N);
          permfunction(pottrans[i], Pb[i], Pbinv[i], N);
        }
      }

      donotdelete = 0;
    }
    ;

    // This constructor allows us to compute the permutation functions externally, incase they are to be reused by several potentials (i.e., in the case of homogeneous priors).
    potential(Scalar** pot, Scalar** pottrans, Scalar* pota, Scalar* potb, int** Pa, int** Pb, int** Painv, int** Pbinv, int N, int method) :
      pot(pot), pottrans(pottrans), pota(pota), potb(potb), Pa(Pa), Pb(Pb), Painv(Painv), Pbinv(Pbinv), N(N), method(method)
    {
      donotdelete = 1;
    }
    ;

    ~potential()
    {
      if (donotdelete == 0)
      {
        for (int i = 0; i < N; i++)
        {
          delete[] pot[i];
          delete[] pottrans[i];
          delete[] Pa[i];
          delete[] Pb[i];
          delete[] Painv[i];
          delete[] Pbinv[i];
        }
        delete[] pot;
        delete[] pottrans;
        delete[] pota;
        delete[] potb;
        delete[] Pa;
        delete[] Pb;
        delete[] Painv;
        delete[] Pbinv;
      }
      for (map<int, Scalar*>::iterator it = messagesa.begin(); it != messagesa.end(); it++)
        delete[] it->second;
      for (map<int, Scalar*>::iterator it = messagesb.begin(); it != messagesb.end(); it++)
        delete[] it->second;
    }
    ;

    void receive(int fromid, int direction, Scalar* message);
    Scalar* send(int toid, int direction, long long& nread);
    int MAP(int direction, long long& nread);
};

#endif
