#ifndef SORT_HPP
#define SORT_HPP

#include "common.hpp"

/**
 A fast sorting algorithm.
 Courtesy of http://warp.povusers.org/SortComparison/
 */

template<typename ItemType>
void InsertionSort(ItemType* array, unsigned size, int* pa)
{
  for (unsigned i = 1; i < size; ++i)
  {
    ItemType val = array[i];
    int vali = pa[i];
    unsigned j = i;
    while (j > 0 && val COMP array[j - 1])
    {
      array[j] = array[j - 1];
      pa[j] = pa[j - 1];
      --j;
    }
    array[j] = val;
    pa[j] = vali;
  }
}

template<typename ItemType>
unsigned Partition(ItemType* array, unsigned f, unsigned l, const ItemType& pivot, int* pa)
{
  unsigned i = f - 1, j = l + 1;
  while (true)
  {
    while (pivot COMP array[--j])
      ;
    while (array[++i] COMP pivot)
      ;
    if (i < j)
    {
      ItemType tmp = array[i];
      array[i] = array[j];
      array[j] = tmp;
      int tmpi = pa[i];
      pa[i] = pa[j];
      pa[j] = tmpi;
    }
    else
      return j;
  }
}

template<typename ItemType>
void QuickSortImpl(ItemType* array, unsigned f, unsigned l, int* pa)
{
  while (f < l)
  {
    unsigned m = Partition(array, f, l, array[f], pa);
    QuickSortImpl(array, f, m, pa);
    f = m + 1;
  }
}

template<typename ItemType>
void QuickSort(ItemType* array, unsigned size, int* pa)
{
  QuickSortImpl(array, 0, size - 1, pa);
}

template<typename ItemType>
void MedianHybridQuickSortImpl(ItemType* array, unsigned f, unsigned l, int* pa)
{
  while (f + 16 < l)
  {
    ItemType v1 = array[f], v2 = array[l], v3 = array[(f + l) / 2];
    ItemType median = v1 < v2 ? (v3 < v1 ? v1 : std::min(v2, v3)) : (v3 < v2 ? v2 : std::min(v1, v3));
    unsigned m = Partition(array, f, l, median, pa);
    MedianHybridQuickSortImpl(array, f, m, pa);
    f = m + 1;
  }
}

template<typename ItemType>
void MedianHybridQuickSort(ItemType* array, unsigned size, int* pa)
{
  MedianHybridQuickSortImpl(array, 0, size - 1, pa);
  InsertionSort(array, size, pa);
}

#endif
